#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <conio.h>
#include <dos.h>
#include <alloc.h>
#include <time.h>

#include "xlib.h"
#include "xpoint.h"
#include "xrect.h"
#include "xline.h"
#include "xtext.h"

#include "mouse.h"

#define ADENINE		99	  // Nucleic acid definitions
#define THYMINE		72
#define CYTOSINE	66
#define GUANINE		58

unsigned char code[1000];	// Genetic code
char invis = 0;				// Is the helix invisible?
char move = 1;				// Should the DNA move around by itself?

void load_pcx(void);

void create_points(void);
void destroy_points(void);

void calc_points(void);
void display_points(void);
void display_info(void);
void draw_dot(int x, int y, unsigned char color);

void move_shape(void);
void control_shape(void);

void handle_input(void);

void load_font(void);

char load_pic = 0;
char pic_file[80];

int bond_frequency = 4;
int num_bonds = 0;

unsigned int num_points;
unsigned int actuelle;
int xangle = 0;
int yangle = 0;
int zangle = 0;
int xdelta = 1;
int ydelta = 1;
int zdelta = -2;

//int object_x = -290;
int object_x = 0;
int object_y = 0;

char direction = 1;

signed int *ox, *oy, *oz;
signed int *sx, *sy, *sz;
signed int *color;
float sine[360];
float cosine[360];

char exit_program = 0;

int scr_width = 376;
int scr_height = 282;
int half_scr_width, half_scr_height;

int main(int argc, char **argv) {

	randomize();

	if (argc > 1) {
		strcpy(pic_file, argv[1]);
		load_pic = 1;
	}

	load_font();

	ms_reset();
	if (mouse_present)
		ms_hide_cursor();

	half_scr_width = scr_width / 2;
	half_scr_height = scr_height / 2;

	x_set_mode(X_MODE_376x282, scr_width);
	x_set_doublebuffer(scr_height);
	create_points();

	while (!exit_program) {
		if (kbhit())
			handle_input();
		move_shape();
		control_shape();
		if (xangle < 0)
			xangle += 359;
		if (yangle < 0)
			yangle += 359;
		if (zangle < 0)
			zangle += 359;
		xangle %= 360;
		yangle %= 360;
		zangle %= 360;

		x_rect_fill(0, 0, scr_width, scr_height, HiddenPageOffs, 0);
		calc_points();
		display_points();
		display_info();
		x_page_flip(0, 0);
	}

	while (kbhit())
		getch();

	x_text_mode();

	destroy_points();
	printf("%d points were rotated in realtime.\n", num_points);
	printf("(Really, %d!)\n", actuelle);
	return 0;
}

	signed int radius = 35;
	int num_circles = 2;
	float squishiness = 1.5;

void create_points(void) {
	signed int i, x, y, a, b, c;
	signed int radius_squared;
	unsigned int p, acid;

	if (!load_pic) {
// Create a random, valid genetic code
		for (i = 0; i < 1000; i+=2) {
			acid = rand() % 4;
			switch (acid) {
				case 0:  code[i] = ADENINE;  code[i+1] = THYMINE;  break;
				case 1:  code[i] = THYMINE;  code[i+1] = ADENINE;  break;
				case 2:  code[i] = CYTOSINE; code[i+1] = GUANINE;  break;
				case 3:  code[i] = GUANINE;  code[i+1] = CYTOSINE; break;
			}
		}

// Create the 3D DNA model
		radius_squared = radius * radius;
//		num_points = 120 + (((radius * 4) * num_circles) * 2);	// with axes
		num_points = ((((radius * 2) + 1) * 4) * num_circles);
		if ((long)((long)num_points * (long)sizeof(int)) > 65535L) {
			x_text_mode();
			printf("%d points is too many!\n", num_points);
			exit(3);
		}
		ox = (int *) malloc(num_points * sizeof(int));
		oy = (int *) malloc(num_points * sizeof(int));
		oz = (int *) malloc(num_points * sizeof(int));

		sx = (int *) malloc(num_points * sizeof(int));
		sy = (int *) malloc(num_points * sizeof(int));
		sz = (int *) malloc(num_points * sizeof(int));

		color = (int *) malloc(num_points * sizeof(int));

		p = 0;
		a = num_circles * 40;
		b = num_circles * 40;

		for (c = 0; c < num_circles; c++) {
			for (x = -radius; x <= radius; x++) {	// 1st half of 1st circle
				ox[p] = x;
				oy[p] = sqrt(radius_squared - (x * x));
				oz[p] = a / squishiness;
				color[p] = 7;
				p++;
				a--;
			}
			for (x = radius; x >= -radius; x--) {	// 2nd half of 1st circle
				ox[p] = x;
				oy[p] = sqrt(radius_squared - (x * x)) * -1;
				oz[p] = a / squishiness;
				color[p] = 7;
				p++;
				a--;
			}
			for (x = radius; x >= -radius; x--) {	// 1st half of 2nd circle
				ox[p] = x;
				oy[p] = sqrt(radius_squared - (x * x)) * -1;
				oz[p] = b / squishiness;
				color[p] = 7;
				p++;
				b--;
			}
			for (x = -radius; x <= radius; x++) {	// 2nd half of 2nd circle
				ox[p] = x;
				oy[p] = sqrt(radius_squared - (x * x));
				oz[p] = b / squishiness;
				color[p] = 7;
				p++;
				b--;
			}
		}
// Draw the axes

/*		for (x = -20; x < 20; x++) {
			ox[p] = x * 2;
			oy[p] = 0;
			oz[p] = 0;
			color[p] = 10;
			p++;
		}
		for (x = -20; x < 20; x++) {
			ox[p] = 0;
			oy[p] = x * 2;
			oz[p] = 0;
			color[p] = 11;
			p++;
		}
		for (x = -20; x < 20; x++) {
			ox[p] = 0;
			oy[p] = 0;
			oz[p] = x * 2;
			color[p] = 12;
			p++;
		} */

/*		for (y = -20; y < 20; y+=4)			// Create a striped square
			for (x = -20; x < 20; x++) {
				ox[p] = x * 4;
				oy[p] = y * 4;
				oz[p] = 0;
				color[p] = 15 + (rand() % 16);
				p++;
			} */

/*		for (i = -200; i < 200; i++) {		// Create stars
			ox[p] = i;
			oy[p] = (rand() % 400) - 200;
			oz[p] = 0;
			color[p] = rand() % 256;
			p++;
		} */
		actuelle = p;
	}
	else
		load_pcx();

	for (i = 0; i < 360; i++) {
		cosine[i] = cos(i * 3.14159265 / 180);
		sine[i] = sin(i * 3.14159265 / 180);
	}
}

void calc_points(void) {
	int i, nx, ny, nz, scr_z;

//// Rotation
	for (i = 0; i < num_points; i++) {
		// X-axis
		ny=(int)(oy[i] * cosine[xangle] - oz[i] * sine[xangle]);
		nz=(int)(oy[i] * sine[xangle] + oz[i] * cosine[xangle]);
		nx=(int)(ox[i]);
		sx[i] = nx;
		sy[i] = ny;

		// Y-axis
		nx=(int)(sx[i] * cosine[yangle] + nz * sine[yangle]);
		nz=(int)(-sx[i] * sine[yangle] + nz * cosine[yangle]);
		sx[i] = nx;

		// Z-axis
		nx=(int)(sx[i] * cosine[zangle] - sy[i] * sine[zangle]);
		ny=(int)(sx[i] * sine[zangle] + sy[i] * cosine[zangle]);
		sx[i] = nx;
		sy[i] = ny;
	}

//// Translation
	for (i = 0; i < num_points; i++) {
		sx[i] = sx[i] + half_scr_width + object_x;
		sy[i] = (0 - sy[i]) + half_scr_height + object_y;
	}
}

void display_points(void) {
	int i;
	int diameter;
	int from, to;
	int midx, midy;
	int cp;						// position in genetic code
	int yeah = 0;

	diameter = (radius * 2) + 1;
	from = 0;
	to = diameter * 2;
	cp = 0;

	i = 0;
	if (bond_frequency > 0)
		if (bond_frequency % 2 == 1)
			bond_frequency++;
	if (bond_frequency > 0)
	while (to < num_points) {
		if (yeah == 0) {
			midx = (sx[from] + sx[to]) / 2;
			midy = (sy[from] + sy[to]) / 2;
			x_line(sx[from], sy[from], midx, midy, code[cp++], HiddenPageOffs);
			x_line(midx, midy, sx[to], sy[to], code[cp++], HiddenPageOffs);
			draw_dot(midx, midy, 15);				// Hydrogen bond
//			draw_dot(sx[from], sy[from], 13);		// Sugar
//			draw_dot(sx[to], sy[to], 13);			// Sugar
			i++;
		}
		else
		if (yeah == bond_frequency / 2) {
//			draw_dot(sx[from], sy[from], 10);		// Phosphate
//			draw_dot(sx[to], sy[to], 10);			// Phosphate
		}
		yeah++;
		yeah %= bond_frequency;

		from += 2;
		to += 2;
		if (from % (diameter * 2) == 0) {
			from += diameter * 2;
			to += diameter * 2;
		}
	}
	num_bonds = i;

	if (!invis)
		for (i = 0; i < num_points; i++) {
			if (sx[i] >= 0 && sx[i] < scr_width && sy[i] >= 0 && sy[i] < scr_height)
				x_put_pix(sx[i], sy[i], HiddenPageOffs, color[i]);
		}
}

	typedef struct {
		char manufacturer;	 	// 10 = ZSoft PCX
		char version;
		char encoding;
		char bits_per_pixel;	// 1, 2, 4, or 8
		int x1, y1, x2, y2;
		int hdpi;
		int vdpi;
		char colormap[48];
		char reserved;			// should be set to 0
		char nplanes;			// # of color planes
		int width;
		int paletteinfo;
		int hscreensize;
		int vscreensize;
		char filler[54];
	} pcx_header;

void load_pcx(void) {
	FILE *pcx;
	unsigned char data;
	long count, num_bytes;
	int i, x = 0, y = 0, p = 0;
	long total_size;
	int xsize, ysize;
	pcx_header header;
	unsigned char r, g, b;

	pcx = fopen(pic_file, "rb");

	fseek(pcx, -768L, SEEK_END);
	for (count = 0; count < 256; count++) {		// Read in the palette
		r = fgetc(pcx) >> 2;
		g = fgetc(pcx) >> 2;
		b = fgetc(pcx) >> 2;
		outp(0x03C8, count);
		outp(0x03C9, r);
		outp(0x03C9, g);
		outp(0x03C9, b);
	}
	fseek(pcx, 0L, SEEK_SET);

	fread(&header, sizeof(header), 1, pcx);
	xsize = header.width;
	ysize = header.y2 - header.y1;

	total_size = (long) xsize * (long) ysize;

	num_points = xsize * ysize;

	ox = (int *) malloc(num_points * sizeof(int));
	if (!ox) {
		x_text_mode();
		printf("Not enough memory for a picture that size!\n");
		exit(1);
	}
	oy = (int *) malloc(num_points * sizeof(int));
	if (!oy) {
		x_text_mode();
		printf("Not enough memory for a picture that size!\n");
		exit(1);
	}
	oz = (int *) malloc(num_points * sizeof(int));
	if (!oz) {
		x_text_mode();
		printf("Not enough memory for a picture that size!\n");
		exit(1);
	}

	sx = (int *) malloc(num_points * sizeof(int));
	if (!sx) {
		x_text_mode();
		printf("Not enough memory for a picture that size!\n");
		exit(1);
	}
	sy = (int *) malloc(num_points * sizeof(int));
	if (!sy) {
		x_text_mode();
		printf("Not enough memory for a picture that size!\n");
		exit(1);
	}
	sz = (int *) malloc(num_points * sizeof(int));
	if (!sz) {
		x_text_mode();
		printf("Not enough memory for a picture that size!\n");
		exit(1);
	}

	color = (int *) malloc(num_points * sizeof(int));

	count = 0;
	while (count <= total_size) {
		data = fgetc(pcx);
		if (data >= 192) {
			num_bytes = data - 192;
			data = fgetc(pcx);
			while (num_bytes-- > 0) {
				x++;
				if (x >= xsize) {
					x = 0;
					y++;
				}
				count++;
				ox[p] = x - (xsize / 2);
				oy[p] = y - (ysize / 2);
				oz[p] = 0;
				color[p++] = data;
//				if (x < xsize)
///					if (video_type == V_MODEX)
///						x_put_pix(x, y, VisiblePageOffs, data);
///					else
///						putPixel(x, y, data);
//				buffer[count++] = data;
			}
		}
		else {
			x++;
			if (x >= xsize) {
				x = 0;
				y++;
			}
			count++;
			ox[p] = x - (xsize / 2);
			oy[p] = y - (ysize / 2);
			oz[p] = 0;
			color[p++] = data;
//			if (x < xsize)
///				if (video_type == V_MODEX)
///					x_put_pix(x, y, VisiblePageOffs, data);
///				else
///					putPixel(x, y, data);
//			buffer[count++] = data;
		}
	}
	fclose(pcx);
}

void move_shape(void) {

	xangle += xdelta;
	yangle += ydelta;
	zangle += zdelta;
	if (move) {
		if (direction == 0) {
			object_x -= 1;
			if (object_x < -75) {
				direction = 1;
				zdelta *= -1;
			}
		}
		else {
			object_x += 1;
			if (object_x > 75) {
				direction = 0;
				zdelta *= -1;
			}
		}
	}
}

void control_shape(void) {
	int mb, mx, my;

	if (mouse_present) {
		ms_read_motion_counters(&mx, &my);
		xangle += (my / 2);
		yangle += (mx / 2);

		mb = ms_get_mouse_pos(&mx, &my);
		if (mb == 1)
			zangle -= 2;
		if (mb == 2)
			zangle += 2;
	}
}

void handle_input(void) {
	char ch;
	int p;
	int from, to, diameter;

	ch = getch();
	if (kbhit()) {		// extended key
		ch = getch();
	}
	else {				// normal key
		switch (ch) {
			case ']': {
				destroy_points();
				squishiness += .10;
				create_points();
				break;
			}
			case '[': {
				destroy_points();
				squishiness -= .10;
				if (squishiness < 0)
					squishiness = 0;
				create_points();
				break;
			}
			case '+': {
				destroy_points();
				num_circles++;
				create_points();
				break;
			}
			case '-': {
				destroy_points();
				if (num_circles > 0)
					num_circles--;
				create_points();
				break;
			}
			case '.': {
				destroy_points();
				radius += 5;
				create_points();
				break;
			}
			case ',': {
				destroy_points();
				if (radius > 0)
					radius -= 5;
				create_points();
				break;
			}
			case '\'': {
				destroy_points();
				if (bond_frequency > 0)
					bond_frequency -= 2;
				create_points();
				break;
			}
			case ';': {
				destroy_points();
				bond_frequency += 2;
				create_points();
				break;
			}
			case 'i': {	  // toggle helix invisibility (so you can see bonds)
				if (invis)
					invis = 0;
				else
					invis = 1;
				break;
			}
			case 'm': {   // toggle automatic DNA movement
				if (move)
					move = 0;
				else
					move = 1;
				break;
			}
			case 'c': {	  // just recreate the model with a new genetic code
				destroy_points();
				create_points();
				break;
			}
			case 'd': {
				x_rect_fill(0, 0, scr_width, scr_height, VisiblePageOffs, 0);
				for (p = 0; p < num_points; p++) {
					x_rect_fill(0, 0, 75, 40, VisiblePageOffs, 0);
					x_printf(0, 0, VisiblePageOffs, 15, "p: %d", p);
					x_put_pix(sx[p], sy[p], VisiblePageOffs, color[p]);
					getch();
				}
				sound(400);
				delay(20);
				sound(600);
				delay(20);
				nosound();
				getch();
				break;
			}
			case '7': xdelta++; break;
			case '1': xdelta--; break;
			case '8': ydelta++; break;
			case '2': ydelta--; break;
			case '9': zdelta++; break;
			case '3': zdelta--; break;
			case '4': xdelta = 0; break;
			case '5': ydelta = 0; break;
			case '6': zdelta = 0; break;
			case ' ': xangle = 0; yangle = 0; zangle = 0; break;
			case 27: exit_program = 1; break;
		}
	}
}

void destroy_points(void) {
	free(ox);
	free(oy);
	free(oz);
	free(sx);
	free(sy);
	free(sz);
	free(color);
}

unsigned char far *newfnt;

void load_font(void) {
	FILE *fp;
	int i;
	char c;

	fp = fopen("dna.fnt", "rb");
	newfnt = (unsigned char far *) farmalloc(256*16+4);
	for (i = 0;i < 256*16+4; i++){
		fread(&c,1,1,fp);
		*(newfnt+i)=c;
	}
	fclose(fp);
	x_text_init();
	x_register_userfont(newfnt);
	x_set_font(2);
}

void display_info(void) {
	int y = 20;
	static signed int ax = 0, tx = 4, cx = 8, gx = 12;
	static char ad = 1, td = 1, cd = 1, gd = 1;

	if (ax > 30)
		ad = 0;
	else
		if (ax < 1)
			ad = 1;
	if (tx > 30)
		td = 0;
	else
		if (tx < 1)
			td = 1;
	if (cx > 30)
		cd = 0;
	else
		if (cx < 1)
			cd = 1;
	if (gx > 30)
		gd = 0;
	else
		if (gx < 1)
			gd = 1;

	if (ad)
		ax++;
	else
		ax--;
	if (td)
		tx++;
	else
		tx--;
	if (cd)
		cx++;
	else
		cx--;
	if (gd)
		gx++;
	else
		gx--;

	x_printf(0, 0, HiddenPageOffs, 15, "Nucleic Acids:");
	x_printf(scr_width / 2, 0, HiddenPageOffs, 15, "Number of bonds: %d", num_bonds);

	x_printf(ax, y, HiddenPageOffs, ADENINE, "Adenine");
	x_printf(tx, y += CharHeight, HiddenPageOffs, THYMINE, "Thymine");
	x_printf(cx, y += CharHeight, HiddenPageOffs, CYTOSINE, "Cytosine");
	x_printf(gx, y += CharHeight, HiddenPageOffs, GUANINE, "Guanine");
}

void draw_dot(int x, int y, unsigned char color) {

	x_put_pix(x,   y, HiddenPageOffs, color);

	x_put_pix(x-1, y-1, HiddenPageOffs, color);
	x_put_pix(x,   y-1, HiddenPageOffs, color);
	x_put_pix(x+1, y-1, HiddenPageOffs, color);
	x_put_pix(x-1, y, HiddenPageOffs, color);
	x_put_pix(x+1, y, HiddenPageOffs, color);
	x_put_pix(x-1, y+1, HiddenPageOffs, color);
	x_put_pix(x,   y+1, HiddenPageOffs, color);
	x_put_pix(x+1, y+1, HiddenPageOffs, color);
}
